"""Functionality related to Python's import mechanisms.
"""

import contextlib
from importlib.abc import MetaPathFinder
from importlib.machinery import ModuleSpec
import sys


class ASTLoader:  # pylint:disable=old-style-class,too-few-public-methods
    """An `importlib.abc.Loader` which loads an AST for a particular name.

    You construct this with an AST and a module name. The
    `exec_module` method simply compiles the AST with the name and
    execs the resulting code against the procided module dict.

    In practice, this is how cosmic-ray loads mutated ASTs for modules.
    """
    def __init__(self, ast, name):
        self._ast = ast
        self._name = name

    def exec_module(self, mod):
        exec(compile(self._ast, self._name, 'exec'),  # pylint:disable=exec-used
             mod.__dict__)


class ASTFinder(MetaPathFinder):  # pylint:disable=too-few-public-methods
    """An `importlib.ast.MetaPathFinder` that associates a module name
    with an AST.

    Construct this by passing a module name and associated AST. When
    the finder is asked for that name, it will return a loader that
    produces the code equivalent of the AST.

    We use this to inject mutated ASTs into tests.
    """
    def __init__(self, fullname, ast):
        self._fullname = fullname
        self._ast = ast

    def find_spec(self, fullname, path, target=None):  # pylint:disable=unused-argument
        if fullname == self._fullname:
            return ModuleSpec(fullname,
                              ASTLoader(self._ast, fullname))
        else:
            return None


@contextlib.contextmanager
def using_mutant(module_name, mutant):
    """Create a new Finder as a context-manager.

    This installs the finder for the duration of the with-block,
    yielding the Finder in the with-statement. After the with-block,
    the Finder is uninstalled.

    This also removes `module_name` from `sys.modules` before doing
    anything else. Then ensures that the finder gets asked to load the
    mutated AST.
    """
    if module_name in sys.modules:
        del sys.modules[module_name]

    finder = ASTFinder(module_name, mutant)
    sys.meta_path = [finder] + sys.meta_path
    try:
        yield finder
    finally:
        sys.meta_path.remove(finder)



import ast
import copy
import itertools

from ..util import get_line_number


def _full_module_name(obj):
    return '{}.{}'.format(
        obj.__class__.__module__,
        obj.__class__.__name__)


class Operator(ast.NodeTransformer):
    """A base class for all mutation operators.

    This takes care of the basic book-keeping that all operators need
    to do. All operators *must* derive from this class since that's
    how we keep track of them.
    """
    def __init__(self, target):
        self._target = target
        self._count = 0
        self._activation_record = None

    @property
    def activation_record(self):
        """The activation record for the operator.

        The activation record is a dict describing where and how the
        operator was applied.
        """
        return self._activation_record

    def visit_mutation_site(self, node):
        """Potentially mutate `node`, returning the mutated version.

        Subclasses should call this when their AST iteration reaches a
        potential mutation site. If that site is scheduled for
        mutation, the subclass instance will be asked to perform the
        mutation.
        """
        if self._count == self._target:
            self._activation_record = {
                'operator': _full_module_name(self),
                'description': str(self),
                'line_number': get_line_number(node)
            }

            old_node = node
            node = self.mutate(old_node)
            ast.copy_location(node, old_node)

        self._count += 1
        return node

    def mutate(self, node):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return `None` if the node has
        been deleted. Return `node` if there is no mutation at all for
        some reason.
        """
        raise NotImplementedError(
            'Mutation operators must implement "mutate()".')

    @classmethod
    def bombard(cls, node):
        """Bombard `node` with cosmic rays, generating a sequence of odious
        mutants.

        The returns an iterable sequence of mutated copies of `node`,
        with one mutant for each potential mutation site in `node`
        with respect to the `Operator` subclass which is performing
        the mutations.
        """
        for target in itertools.count():
            operator = cls(target)
            mutant = operator.visit(copy.deepcopy(node))
            if not operator.activation_record:
                break

            yield operator.activation_record, mutant



import importlib
import logging
import pkgutil

LOG = logging.getLogger()


def find_modules(name):
    """Generate sequence of all submodules of NAME, including NAME itself.

    Given a directory structure like this:

        /a/
          __init__.py
          b.py
          c/
              __init__.py
              d.py

    you get this:

        >>> list(find_modules('a'))
        [<module 'a' from 'a/__init__.py'>,
         <module 'a.b' from 'a/b.py'>,
         <module 'a.c' from 'a/c/__init__.py'>,
         <module 'a.c.d' from 'a/c/d.py'>]
    """
    module_names = [name]
    while module_names:
        module_name = module_names.pop()
        try:
            module = importlib.import_module(module_name)

            yield module

            if hasattr(module, '__path__'):
                for _, name, _ in pkgutil.iter_modules(module.__path__):
                    module_names.append(
                        '{}.{}'.format(
                            module_name, name))
        except Exception:  # pylint:disable=broad-except
            LOG.exception(
                'Unable to import %s',
                module_name)



from itertools import chain
import unittest

from .test_runner import TestRunner


class UnittestRunner(TestRunner):  # pylint:disable=no-init, too-few-public-methods
    """A TestRunner using `unittest`'s discovery mechanisms.

    This discovers all tests under `test_dir` and executes them.
    """

    def _run(self):
        suite = unittest.TestLoader().discover(self.test_dir)
        result = unittest.TestResult()
        result.failfast = True
        suite.run(result)

        return (
            result.wasSuccessful(),
            [(str(r[0]), r[1])
             for r in chain(result.errors,
                            result.failures)])



"""cosmic-ray

Usage:
  cosmic-ray run [options] [--exclude-modules=P ...] <top-module> <test-dir>
  cosmic-ray load <config-file>

Options:
  -h --help           Show this screen.
  --timeout=T         Maximum time (seconds) a mutant may run [default: 5]
  --verbose           Produce verbose output
  --no-local-import   Allow importing module from the current directory
  --test-runner=R     Test-runner plugin to use [default: unittest]
  --exclude-modules=P Pattern of module names to exclude from mutation
"""

import docopt


def load_file(config_file):
    """Read configuration from a file.

    This reads `config_file`, yielding each non-empty line with
    whitespace and comments stripped off.
    """
    with open(config_file, 'rt', encoding='utf-8') as f:
        for line in f:
            line = line.split('#')[0] # Remove comments
            line = line.strip()       # Remove edge-whitespace
            if line:
                yield line


def load_configuration():
    config = docopt.docopt(__doc__, version='cosmic-ray v.2')

    if config['load']:
        filename = config['<config-file>']
        args = list(load_file(filename))
        config = docopt.docopt(__doc__, argv=args, version='cosmic-ray v.2')

    return config



import logging
import multiprocessing
import re
import sys

from stevedore import driver, extension

from cosmic_ray.config import load_configuration
from cosmic_ray.find_modules import find_modules
from cosmic_ray.mutating import create_mutants, run_with_mutant
import cosmic_ray.operators
from cosmic_ray.testing.test_runner import TestResult, Outcome


LOG = logging.getLogger()


def format_test_result(mutation_record, test_result):
    """Returns a reasonably formatted string with test outcome,
    activation-record information, and reason.
    """
    arec = mutation_record.activation_record

    return '{outcome} -> {desc} @ {filename}:{lineno}\n{reason}'.format(
        outcome=test_result.outcome,
        desc=arec['description'],
        filename=mutation_record.module_file,
        lineno=arec['line_number'],
        reason=test_result.results)


def hunt(mutation_records, test_runner, timeout):
    """Call `test_runner` for each mutant in `mutation_records`.

    `test_runner` should be a `TestRunner` instance.

    Returns a sequence of `(MutationRecord, TestResult)` tuples.
    """

    with multiprocessing.Pool(maxtasksperchild=1) as pool:
        test_results = ((rec,
                         pool.apply_async(run_with_mutant,
                                          args=(test_runner, rec)))
                        for rec in mutation_records)

        LOG.info('all tests initiated')

        for rec, async_result in test_results:
            try:
                # TODO: This timeout needs to be configurable.
                result = async_result.get(timeout=timeout)
            except multiprocessing.TimeoutError:
                result = TestResult(Outcome.INCOMPETENT, 'timeout')

            LOG.info('mutation record: %s', rec)
            LOG.info('result: %s', result)

            yield (rec, result)


def filtered_modules(modules, excludes):
    exclude_patterns = [re.compile(ex) for ex in excludes]
    for module in modules:
        if not any([pattern.match(module.__name__)
                    for pattern in exclude_patterns]):
            yield module


def main():
    configuration = load_configuration()

    if configuration['--verbose']:
        logging.basicConfig(level=logging.INFO)

    timeout = float(configuration['--timeout'])

    if not configuration['--no-local-import']:
        sys.path.insert(0, '')

    modules = filtered_modules(
        find_modules(configuration['<top-module>']),
        configuration['--exclude-modules'])

    # We don't use the extensions directly. This just forces importing
    # of modules which contain operators.
    extension.ExtensionManager(
        namespace='cosmic_ray.operators')

    operators = cosmic_ray.operators.all_operators()

    test_runner = driver.DriverManager(
        namespace='cosmic_ray.test_runners',
        name=configuration['--test-runner'],
        invoke_on_load=True,
        invoke_args=(configuration['<test-dir>'],),
    ).driver

    results = hunt(
        mutation_records=create_mutants(modules, operators),
        test_runner=test_runner,
        timeout=timeout)

    outcomes = {o: 0 for o in Outcome}

    for mutation_record, test_result in results:
        outcomes[test_result.outcome] += 1
        print(format_test_result(mutation_record, test_result))  # pylint:disable=superfluous-parens

    total_count = sum(outcomes.values())

    if total_count > 0:
        print('Survival rate: {:0.2f}%'.format(
            100 * outcomes[Outcome.SURVIVED] / total_count))
    else:
        print('No tests run (no mutations generated).')  # pylint:disable=superfluous-parens

    sys.exit(0 if outcomes[Outcome.SURVIVED] == 0 else 1)

if __name__ == '__main__':
    main()



import ast

from .operator import Operator


class NumberReplacer(Operator):
    """An operator that modifies numeric constants.
    """

    def visit_Num(self, node):  # noqa
        return self.visit_mutation_site(node)

    def mutate(self, node):
        """Modify the numeric value on `node`.
        """
        new_node = ast.Num(n=node.n + 1)
        return new_node

    def __repr__(self):
        return 'NumberReplacer(target={})'.format(
            self._target)



"""This module contains mutation operators which replace one
relational operator with another.

For each relational operator we generate a series of classes, one for
each *other* operator. These generated classes mutate nodes by
replaces the first operator with the second.

The generated classes are named `Replace<from-op>With<to-op>` where
`from-op` is the relational operator being replaced and `to-op` is the
replacement operator.

The global list `operators` also contains a all of the generated
classes. This list is primarily used for testing.
"""

import ast

from .operator import Operator


RELATIONAL_OPERATORS = {ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE,
                        ast.Is, ast.IsNot, ast.In, ast.NotIn}

# There are a number of potential replacements which we avoid because
# they almost always produce equivalent mutants. This is a set of
# (FROM-OP, TO-OP) tuples which we don't want to generate.
SKIP = {
    (ast.NotEq, ast.IsNot),
}


def create_operator(from_op, to_op):  # pylint:disable=redefined-outer-name
    """Create an operator which replaces `from_op` with `to_op`.

    This inserts the class into the global namespace with the name
    `ReplaceXXXWithYYY'.

    Returns the new class
    """
    operator_name = 'Replace{}With{}'.format(
        from_op.__name__, to_op.__name__)

    visit_func_name = 'visit_{}'.format(from_op.__name__)
    visit_func = lambda self, node: self.visit_mutation_site(node)  # noqa

    new_op = type(
        operator_name,
        (Operator,),
        {'mutate': lambda self, node: to_op(),
         visit_func_name: visit_func,
         'from_op': from_op,
         '__repr__': lambda self: 'replace {} with {}'.format(
             from_op.__name__, to_op.__name__)})

    globals()[operator_name] = new_op
    return new_op


# For each relational operator A, create one class for each *other*
# relational operator B which replaces A with B in an AST.
OPERATORS = [create_operator(from_op, to_op)
             for from_op in RELATIONAL_OPERATORS
             for to_op in RELATIONAL_OPERATORS.difference({from_op})
             if (from_op, to_op) not in SKIP]



import ast

from .operator import Operator


class ReverseUnarySub(Operator):
    """A NodeTransformer that deletes unary subtraction (i.e. negation).
    """

    def visit_UnaryOp(self, node):  # noqa
        if isinstance(node.op, ast.USub):
            return self.visit_mutation_site(node)
        else:
            return node

    def mutate(self, node):
        """Replace the unary-sub operator with unary-add.
        """
        node.op = ast.UAdd()
        return node

    def __repr__(self):
        return 'ReverseUnarySub(target={})'.format(
            self._target)


class ReverseUnaryAdd(Operator):
    """A NodeTransformer that reverses unary addition (i.e. the positive sign).
    """

    def visit_UnaryOp(self, node):  # noqa
        if isinstance(node.op, ast.UAdd):
            return self.visit_mutation_site(node)
        else:
            return node

    def mutate(self, node):
        """Replace the unary-add operator with unary-sub.
        """
        node.op = ast.USub()
        return node

    def __repr__(self):
        return 'ReverseUnaryAdd(target={})'.format(
            self._target)



import abc
from collections import namedtuple
from enum import Enum


class Outcome(Enum):
    """A enum of the possible outcomes for any mutant test run.
    """
    SURVIVED = 'survived'
    KILLED = 'killed'
    INCOMPETENT = 'incompetent'


# The type returned by TestRunner runs.
TestResult = namedtuple('TestResult',
                        ['outcome',
                         'results'])


class TestRunner(metaclass=abc.ABCMeta):
    """Specifies the interface for test runners in the system.

    There are many ways to run unit tests in Python, and each method
    supported by Cosmic Ray should be provided by a TestRunner
    implementation.
    """
    def __init__(self, test_dir):
        self._test_dir = test_dir

    @property
    def test_dir(self):
        """The directory containing the tests to be run.
        """
        return self._test_dir

    @abc.abstractmethod
    def _run(self):
        """Run all of the tests and return the results.

        The results are returned as a (success, result)
        tuple. `success` is a boolean indicating if the tests
        passed. `result` is any object that is appropriate to provide
        more information about the success/failure of the tests.
        """
        raise NotImplemented()

    def __call__(self):
        """Call `_run()` and return a `TestResult` with the results.

        The `outcome` field of the return value is:

        - SURVIVED if the tests passed
        - KILLED if the tests failed
        - INCOMPETENT if the tests raised an exception.

        The `results` field is simply the second field of
        `test_runner.run()`'s return value.
        """
        try:
            test_result = self._run()
            if test_result[0]:
                return TestResult(Outcome.SURVIVED,
                                  test_result[1])
            else:
                return TestResult(Outcome.KILLED,
                                  test_result[1])
        except Exception as e:
            return TestResult(Outcome.INCOMPETENT,
                              str(e))



def get_line_number(node):
    """Try to get the line number for `node`.

    If no line number is available, this returns "<UNKNOWN>".
    """
    if hasattr(node, 'lineno'):
        return node.lineno
    else:
        return '<UNKNOWN>'



import ast
import logging
from collections import namedtuple

from .importing import using_mutant


LOG = logging.getLogger()


MutationRecord = namedtuple('MutationRecord', ['module_name',
                                               'module_file',
                                               'operator',
                                               'activation_record',
                                               'mutant'])


def create_mutants(modules, operators):
    """Mutate `modules` with `operators`.

    Returns an iterable of `MutationRecord`s, one for each application
    of an operator to a location in a module.

    """
    for module in modules:
        with open(module.__file__, 'rt', encoding='utf-8') as module_file:
            LOG.info('reading module %s from %s',
                     module.__name__,
                     module.__file__)
            source = module_file.read()

        pristine_ast = ast.parse(source, module.__file__, 'exec')

        for operator in operators:
            for activation_record, mutant in operator.bombard(pristine_ast):
                yield MutationRecord(module_name=module.__name__,
                                     module_file=module.__file__,
                                     operator=operator,
                                     activation_record=activation_record,
                                     mutant=mutant)


def run_with_mutant(func, mutation_record):
    """Install the mutation record and run the callable `func`, returning
    its result.
    """
    module_name, _, _, _, mutant = mutation_record
    with using_mutant(module_name, mutant):
        return func()



from setuptools import setup, find_packages

with open('README.md', 'rt') as readme:
    long_description = readme.read()

setup(
    name='cosmic_ray',
    version='0.1.1',
    packages=find_packages(),

    author='Sixty North AS',
    author_email='austin@sixty-north.com',
    description='Mutation testing',
    license='MIT License',
    keywords='testing',
    url = 'http://github.com/sixty-north/cosmic-ray',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],
    platforms='any',
    include_package_data=True,
    install_requires=[
        'decorator',
        'docopt',
        'stevedore',
        'with_fixture',
    ],
    entry_points={
        'console_scripts': [
            'cosmic-ray = cosmic_ray.app:main',
        ],
        'cosmic_ray.test_runners': [
            'unittest = cosmic_ray.testing.unittest_runner:UnittestRunner',
        ],
        'cosmic_ray.operators': [
            'number_replacer = '
            'cosmic_ray.operators.number_replacer:NumberReplacer',
            'relational_operator_replacement ='
            'cosmic_ray.operators.relational_operator_replacement:create_operator',
            'arithmetic_operator_deletion ='
            'cosmic_ray.operators.arithmetic_operator_deletion:ReverseUnarySub',
        ],
    },
    long_description=long_description
)
